/*
 * new transdisk by Toni Wilen <twilen@arabuusimiehet.com>
 *
 * New features compared to old transdisk:
 * - raw mode (UAE 0.8.13 or later required, preferably 0.8.18 or later)
 *
 * RAW mode disables interrupts and uses direct hardware access. Unfortunately
 * there isn't system routine for reading multiple disk revolutions with accurate
 * index syncs. (No, TD_RAWREAD isn't enough)
 * Tested with Kickstart 1.3 (Basic A500) and 3.0 (A1200+50MHz 68030)
 *
 * NOTE: THIS PROGRAM CAN'T BE USED COPYING PROTECTED DISKS!
 * 
 * This program can create images of copy protected Amiga disks, but it is impossible
 * to write them back to disk with standard Amiga hardware. Even with special hardware it
 * requires careful analyzation of track data and knowledge of specific protection
 * scheme. Emulating image file as floppy is very different to writing to real floppy!
 * DO NOT ASK THE AUTHOR ABOUT COPY PROTECTIONS. This program "just" creates
 * identical image of MFM data from floppy, nothing more, nothing less..
 *
 * In theory this program can create images from all MFM (2us bit rate) formatted
 * floppies except floppies with bit cell timing tricks.
 *
 * Tested with following original games:
 * - Federation of Free Traders
 * - MidWinter
 * - Time Machine
 * Also works with MS-DOS and diskspare formatted floppies
 *
 * Changelog:
 *
 * 05.12.2000
 * - first version
 * 03.04.2001
 * - rewritten analyzer, works now with MS-DOS formatted floppies
 * 02.12.2001
 * - fixed "empty track"-bug
 *
 *
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <exec/types.h>
#include <exec/memory.h>
#include <exec/devices.h>
#include <clib/exec_protos.h>
#include <clib/alib_protos.h>
#include <clib/dos_protos.h>
#include <devices/trackdisk.h>
#include <libraries/dosextens.h>

#define USE_ASSEMBLY_CODE

/* space allocated for track buffer */
#define LONGEST_TRACK 16000

/* save tracks larger than or equal of this as raw data always, even if they seem
 * to have normal AmigaDOS format (with long GAP). Long track is popular copyprotection
 * method (it is impossible to write long tracks with standard Amiga hardware)
 */
#define LONG_TRACK 12800

#define MAX_CYLINDERS 83
#define MAX_TRACKS 2*MAX_CYLINDERS

typedef enum { TRACK_AMIGADOS, TRACK_RAW } image_tracktype;
typedef enum { ADF_NORMAL, ADF_EXT1, ADF_EXT2 } drive_filetype;

struct trackid {
	int trackbytes;
	int trackbits;
	image_tracktype type;
};
static drive_filetype imagetype;
static int amigatracklength;
static int starttrack,endtrack;
static int drive;
static UBYTE *imagefilename;
static UBYTE conversiontable[MAX_TRACKS];
static UBYTE *writebuffer;
static UWORD *amigamfmbuffer;
static UBYTE *tmpmfmbuffer;
static int romrip;

volatile UBYTE *ciaa=(UBYTE*)0xbfe001;
volatile UBYTE *ciab=(UBYTE*)0xbfd000;
volatile UWORD *custom=(UWORD*)0xdff000;

static struct trackid tracks[MAX_TRACKS];

static struct MsgPort *port;
static struct IOExtTD *req;

static void motor(int moto)
{
	req->iotd_Req.io_Length=moto;
	req->iotd_Req.io_Command=TD_MOTOR;
	DoIO((struct IORequest*)req);
}

static void seek(int track)
{
	req->iotd_Req.io_Offset=track*amigatracklength;
	req->iotd_Req.io_Command=TD_SEEK;
	DoIO((struct IORequest*)req);
}

static void read(int track,UBYTE *buf)
{
	req->iotd_Req.io_Offset=track*amigatracklength;
	req->iotd_Req.io_Command=CMD_READ;
	req->iotd_Req.io_Data=buf;
	req->iotd_Req.io_Length=amigatracklength;
	DoIO((struct IORequest*)req);
}

void freedrive(void)
{
	if(!req) return;
	motor(0);
	CloseDevice((struct IORequest*)req);
	DeleteExtIO((struct IORequest*)req);
	DeletePort(port);
	req=0;
	port=0;
}

static int allocdrive(void)
{
	port=0;
	req=0;
	port=CreatePort(0,0);
	if(!port) goto error;
	req=(struct IOExtTD*)CreateExtIO(port,sizeof(struct IOExtTD));
	if(!req) goto error;
	if(OpenDevice("trackdisk.device",drive,(struct IORequest*)req,0)) goto error;
	return 1;
	error:
	if(req) DeleteExtIO((struct IORequest*)req);
	if(port) DeletePort(port);
	return 0;
}

static int bitshiftcompare(UBYTE *src,int bit,int len,UBYTE *comp)
{
	UBYTE b;
	int ones,zeros,len2;

	ones=zeros=0;
	len2=len;
	while(len--) {
		b = (comp[0] << bit) | (comp[1] >> (8 - bit));
		if(b != *src) return 1;
		if(b==0x00) zeros++;
		if(b==0xff) ones++;
		src++;
		comp++;
	}
	if(ones==len2||zeros==len2) return 1;
	return 0;
}

static UBYTE *mergepieces(UBYTE *start,int len,int bits,UBYTE *sync)
{
	UBYTE *dst=tmpmfmbuffer;
	UBYTE b;
	int size;
	int shift;
	
	size=len-(sync-start);
	memcpy(dst,sync,size);
	dst+=size;
	b=start[len];
	b&=~(255>>bits);
	b|=start[0]>>bits;
	*dst++=b;
	shift=8-bits;
	while(start<=sync+2000) {
		*dst++=(start[0]<<shift)|(start[1]>>(8-shift));
		start++;
	}
	return tmpmfmbuffer;
}	
		
#define SCANOFFSET 1 /* scanning range in bytes, -SCANOFFSET to SCANOFFSET */
#define SCANOFFSET2 20
#define SCANLENGHT 200 /* scanning length in bytes */

static UBYTE* scantrack(UBYTE *sync1,UBYTE *sync2,int *trackbytes,int *trackbits)
{
	int i,bits,bytes,matched;
	UBYTE *sync2bak=sync2;
	
	sync1+=SCANOFFSET2;
	sync2+=SCANOFFSET2;
	while(sync1 < sync2bak - 2*SCANOFFSET - SCANOFFSET2 - SCANLENGHT) {
		matched=0x7fff;
		for(i=0;i<2*SCANOFFSET*8;i++) {
			bits=i&7;
			bytes=-SCANOFFSET+(i>>3);
			if(!bitshiftcompare(sync1,bits,SCANLENGHT,sync2+bytes)) {
				if(matched==0x7fff) {
					matched=i;
				} else {
					break;
				}
			}
		}
		if(matched!=0x7fff && i>=2*SCANOFFSET*8) {
			bits=matched&7;
			bytes=-SCANOFFSET+(matched>>3);
			*trackbytes=sync2+bytes-sync1;
			*trackbits=bits;
			return mergepieces(sync1,*trackbytes,*trackbits,sync2bak);
		}
		sync1++;
		sync2++;
	}
	return 0;
}					
					
#define ISMOUSE (!(ciaa[0]&0x40))

#ifdef USE_ASSEMBLY_CODE
extern void __asm asmrawread(register __a0 UBYTE*,register __a1 UBYTE**);
#endif

/* read one track as raw MFM data, also mark index sync positions */
static int rawread(UBYTE *buf,UBYTE **syncpos,FILE *f,int track)
{
	UWORD adkcon;
	#ifndef USE_ASSEMBLY_CODE
	int i;
	UWORD v;
	#endif
	UBYTE *ptr1;
	UBYTE b;

	syncpos[0]=syncpos[1]=syncpos[2]=0;
	Disable();
	/* clear DSKLEN */
	custom[0x24>>1]=0x0000;
	/* set ADKCON */
	adkcon=custom[0x10>>1];
	custom[0x9e>>1]=0x7fff;
	custom[0x9e>>1]=0x8000|0x1000|0x0100; /* set MFMPREC and FAST */
	/* select drive (motor is already started) */
	ciab[0x100]|=0x08|0x10|0x20|0x40;
	ciab[0x100]&=~0x80;
	ciab[0x100]&=~(0x08<<drive);
	/* select side */
	if(track&1) ciab[0x100]&=~0x04; else ciab[0x100]|=0x04;
	/* disable diskindex interrupts, normally not set but you never know.. */
	ciab[0xd00]=0x10;
	/* clear cia interrups */
	b=ciab[0xd00];

	ptr1=buf;
#ifdef USE_ASSEMBLY_CODE
	asmrawread(ptr1,syncpos);
#else
	i=0;
	/* wait for index sync */
	while((ciab[0xd00]&0x10)==0);
	/* read 3 revolutions of track */
	for(;;) {
		v=custom[0x1a>>1];
		if(v&0x8000) *ptr1++=v;
		if(ciab[0xd00]&0x10) {
			/* index sync */
			syncpos[i++]=ptr1;
			if(i>3) break;
		}
	}
#endif
	/* deselect drive */
	ciab[0x100]|=(0x08<<drive);
	/* reset ADKCON */
	custom[0x9e>>1]=0x7fff;
	custom[0x9e>>1]=0x8000|adkcon;
	/* clear cia diskindex irq */
	b=ciab[0xd00];
	/* clear EXTER */
	custom[0x9c]=0x2000;
	Enable();
	if(ISMOUSE) return 0;
	return 1;
}

#define MFMMASK 0x55555555
static ULONG getmfmlong (UWORD * mbuf)
{
	return (ULONG)(((*mbuf << 16) | *(mbuf + 1)) & MFMMASK);
}

#define FLOPPY_WRITE_LEN 6250

static int drive_write_adf_amigados (UWORD *mbuf, UWORD *mend, int track)
{
	int i, secwritten = 0;
	ULONG odd, even, chksum, id, dlong;
	UBYTE *secdata;
	UBYTE secbuf[544];
	char sectable[11];

	memset (sectable, 0, sizeof (sectable));
	mend -= (4 + 16 + 8 + 512);
	while (secwritten < 11) {
		int trackoffs;

		do {
			while (*mbuf++ != 0x4489) {
				if (mbuf >= mend)
				return 1;
			}
		} while (*mbuf++ != 0x4489);

		odd = getmfmlong (mbuf);
		even = getmfmlong (mbuf + 2);
		mbuf += 4;
		id = (odd << 1) | even;

		trackoffs = (id & 0xff00) >> 8;
		if (trackoffs > 10) return 2;
		chksum = odd ^ even;
		for (i = 0; i < 4; i++) {
			odd = getmfmlong (mbuf);
			even = getmfmlong (mbuf + 8);
			mbuf += 2;

			dlong = (odd << 1) | even;
			if (dlong) return 6;
			chksum ^= odd ^ even;
		} /* could check here if the label is nonstandard */
		mbuf += 8;
		odd = getmfmlong (mbuf);
		even = getmfmlong (mbuf + 2);
		mbuf += 4;
		if (((odd << 1) | even) != chksum || ((id & 0x00ff0000) >> 16) != track) return 3;
		odd = getmfmlong (mbuf);
		even = getmfmlong (mbuf + 2);
		mbuf += 4;
		chksum = (odd << 1) | even;
		secdata = secbuf + 32;
		for (i = 0; i < 128; i++) {
			odd = getmfmlong (mbuf);
			even = getmfmlong (mbuf + 256);
			mbuf += 2;
			dlong = (odd << 1) | even;
			*secdata++ = dlong >> 24;
			*secdata++ = dlong >> 16;
			*secdata++ = dlong >> 8;
			*secdata++ = dlong;
			chksum ^= odd ^ even;
		}
		mbuf += 256;
		if (chksum) return 4;
		sectable[trackoffs] = 1;
		secwritten++;
		memcpy (writebuffer + trackoffs * 512, secbuf + 32, 512);
	}
    if (secwritten == 0 || secwritten < 0) return 5;
	return 0;
}

/* search and align to 0x4489 WORDSYNC markers */
static UBYTE *isamigatrack(UBYTE *mfmdata, int track)
{
	UWORD *dst = amigamfmbuffer;
	int len = LONGEST_TRACK * 8;
	int shift, syncshift, sync,ret;
	ULONG l;
	UWORD w;

	sync = syncshift = shift = 0;
	while (len--) {
		l = (mfmdata[0] << 16) | (mfmdata[1] << 8) | (mfmdata[2] << 0);
		w = l >> (8 - shift);
		if (w == 0x4489) {
			sync = 1;
			syncshift = 0;
		}
		if (sync) {
			if (syncshift == 0) *dst++ = w;
			syncshift ++;
			if (syncshift == 16) syncshift = 0;
		}
		shift++;
		if (shift == 8) {
			mfmdata++;
			shift = 0;
		}
	}
	if (sync) {
		ret=drive_write_adf_amigados (amigamfmbuffer, dst, track);
		if(!ret) return writebuffer;
	printf(" %d\n",ret);
	}
	return 0;
}


/* read and examine one raw track, do MFM decoding if track
 * is AmigaDOS formatted and track conversion is not disabled
 */
static int dotrack(int track,UBYTE *buf,FILE *f)
{
	int retries,trackbytes,trackbits,type;
	int out = 0;
	UBYTE *syncpos[10]={0},*trackbuffer,*amigatrack;

	trackbuffer=0;
	for(retries=0;retries<3;retries++) {
		if(!rawread(buf,syncpos,f,track)) return 0;
		trackbytes=trackbits=0;
		trackbuffer=scantrack(syncpos[1],syncpos[2],&trackbytes,&trackbits);
		if(trackbuffer) break;
	}
	if(!trackbuffer) {
		trackbuffer=tmpmfmbuffer;
		memset(trackbuffer,0,LONGEST_TRACK);
	}
	if(!trackbytes) {
		printf("empty track?");
		trackbytes=trackbits=0;
	} else {
		printf("length %d/%d",trackbytes,trackbits);
	}

	type=TRACK_RAW;
	amigatrack=0;

	if(!conversiontable[track]&&trackbytes) {
		amigatrack=isamigatrack(trackbuffer, track);
		if(amigatrack) {
			if(trackbytes<LONG_TRACK) {
				trackbuffer=amigatrack;
				trackbytes=11*512;
				trackbits=0;
				type=TRACK_AMIGADOS;
				printf(" (AmigaDOS)");
				out=1;
			} else {
				printf(" (long AmigaDOS)");
				out=1;
			}
		}
	}
	if(!out) printf (" (raw)");

	tracks[track].trackbits=trackbytes*8+trackbits;
	/* clear unneeded bits from last byte */
	trackbuffer[trackbytes]&=~(255>>(8-trackbits));
	trackbuffer[trackbytes+1]=0;
	if(trackbits) trackbytes++;
	/* track length must be word aligned */
	trackbytes=(trackbytes+1)&~1;
	tracks[track].trackbytes=trackbytes;
	if(trackbytes) fwrite(trackbuffer,tracks[track].trackbytes,1,f);
	tracks[track].type=type;
	return 1;
}

/* create normal (non-raw) ADF file */
static int doamigatrack(int track,UBYTE *buf,FILE *f)
{
	read(track,buf);
	fwrite(buf,amigatracklength,1,f);
	if(ISMOUSE) return 0;
	return 1;
}

static void writeword(FILE *f,int w)
{
	UBYTE b;

	b=w>>8;
	fwrite(&b,1,1,f);
	b=w;
	fwrite(&b,1,1,f);
}

static void writelong(FILE *f,int l)
{
	writeword(f,l>>16);
	writeword(f,l);
}

/* update ADF_EXT2 header */
static void fiximage(FILE *f,int numtracks)
{
	int i;

	if(imagetype!=ADF_EXT2) return;
	fseek(f,8+4,SEEK_SET);
	for(i=0;i<numtracks;i++) {
		writelong(f,tracks[i].type);
		writelong(f,tracks[i].trackbytes);
		writelong(f,tracks[i].trackbits);
	}
}

/* create empty ADF_EXT2 */
static void initimage(FILE *f,int numtracks)
{
	if(imagetype!=ADF_EXT2) return;
	fwrite("UAE-1ADF",8,1,f);
	writelong(f,numtracks);
	while(numtracks--) {
		writelong(f,0);
		writelong(f,0);
		writelong(f,0);
		tracks[numtracks].type=TRACK_RAW;
	}
}

static void usage(void)
{
	printf(
		"Usage: rawread [options]\n"
		"Options:\n"
		"-f: image filename, must be specified\n"
		"-u: floppy drive unit, 0-3 (default 0 = DF0)\n"
		"-s n: first cylinder, default 0\n"
		"-e n: last cylinder, default 79, maximum %d\n"
		"-r: raw mode (read non AmigaDOS or copy protected floppies)\n"
		"-x n: write track n as raw, not AmigDOS formatted, -1 = disable all conversions\n"
		"-w: write to floppy (AmigaDOS tracks only)\n"
		"-k: dump kickstart rom\n"
		,MAX_CYLINDERS-1);
}

static int readparams(int argc,char *argv[])
{
	int i,sectors,j;

	sectors=11;
	starttrack=0;
	endtrack=79;
	drive=0;
	imagefilename=0;
	imagetype=ADF_NORMAL;
	for (i = 1; i < argc;) {
		if (argv[i][0] != '-' || argv[i][2] != 0) goto error;
		switch(tolower(argv[i][1]))
			{
			case 'h':
			sectors=22;
			i++;
			break;
			case 's':
			if(i+1>=argc) goto error;
			starttrack=atoi(argv[i+1]);
			if(starttrack<0||starttrack>=MAX_CYLINDERS) goto error;
			i+=2;
			break;
			case 'e':
			if(i+1>=argc) goto error;
			endtrack=atoi(argv[i+1]);
			if(endtrack<0||endtrack>=MAX_CYLINDERS) goto error;
			i+=2;
			break;
			case 'u':
			if(i+1>=argc) goto error;
			drive=atoi(argv[i+1]);
			if(drive<0||drive>3) goto error;
			i+=2;
			break;
			case 'r':
			imagetype=ADF_EXT2;
			i++;
			break;
			case 'x':
			if(i+1>=argc) goto error;
			j=atoi(argv[i+1]);
			if(j<-1||j>=MAX_TRACKS) goto error;
			if(j==-1) {
				for(j=0;j<MAX_TRACKS;j++) conversiontable[j]=1;
			} else {
				conversiontable[j]=1;
			}
			i+=2;
			break;
			case 'f':
			if(i+1>=argc) goto error;
			imagefilename=argv[i+1];
			i+=2;
			break;
			case 'k':
			i++;
			romrip=1;
			break;
			default:
			goto error;
		}
	}
	if(imagefilename==0) goto error;
	if(starttrack>=endtrack) goto error;
	amigatracklength=512*sectors;
	return 1;
	error:
	usage();
	return 0;
}

static int ripdisk(FILE *f)
{	
	UBYTE *buf,*chipbuf;
	int track;

	buf = 0;
	writebuffer = 0;
	amigamfmbuffer = 0;
	tmpmfmbuffer = 0;
	chipbuf = 0;

	buf=malloc(LONGEST_TRACK*4);
	if(!buf) goto error;
	writebuffer = malloc (11 * 512);
	if (!writebuffer) goto error;
	amigamfmbuffer = malloc (LONGEST_TRACK);
	if (!amigamfmbuffer) goto error;
	tmpmfmbuffer = malloc(LONGEST_TRACK);
	if (!tmpmfmbuffer) goto error;

	chipbuf=AllocMem(amigatracklength,MEMF_CHIP);
	if(!chipbuf) goto error;

	if(!allocdrive()) goto error;
	motor(1);

	initimage(f,MAX_TRACKS);
	printf("start cylinder: %d, end cylinder %d (total %d cylinders)\n",starttrack,endtrack,endtrack-starttrack+1);
	starttrack*=2;
	endtrack*=2;
	endtrack++;
	track=starttrack;
	while(track<=endtrack) {
		printf("cyl %2d side %d: ",track/2,track&1);
		if(imagetype==ADF_EXT2) {
			seek(track);
			if(!dotrack(track,buf,f)) break;
		} else {
			if(!doamigatrack(track,chipbuf,f)) break;
		}
		printf("\n");
		fflush(stdout);
		track++;
	}
	printf("\n");
	fiximage(f,MAX_TRACKS);

	error:
	freedrive();
	if (chipbuf) FreeMem(chipbuf,amigatracklength);
	free (writebuffer);
	free (tmpmfmbuffer);
	free (amigamfmbuffer);
	free (buf);

	return 1;
}

static int riprom(FILE *f)
{
	UBYTE *addr1 = (UBYTE*)0xf80000;
	UBYTE *addr2 = (UBYTE*)0xfc0000;
	int len;

	if (!memcmp(addr1,addr2,262144))
		len = 262144;
		else
		len = 524288;
	fwrite (addr1, len, 1, f);
	return 1;
}

int main(int argc,char *argv[])
{
	FILE *f;

	f = 0;

	if(!readparams(argc,argv)) return 1;

	strcat(imagefilename,romrip ? ".rom" : ".adf");
	f=fopen(imagefilename,"wb");
	if(!f) {
		printf("Couldn't open '%s' for writing\n",imagefilename);
		return 0;
	}

	if(romrip) {
		riprom(f);
	} else {
		ripdisk(f);
	}
	
	fclose(f);

	return 0;
}
